function  C = fundamental_cycle_basis(W,root)
%
% C= fundamental_cycle_basis(A,weights,cam)
% This function computes a Fundamental Cycle Basis of an undirected graph,
% i.e. a basis obtained from the cycles induced by spanning tree.
% It is assumed that the underlying graph is connected.
% Let nV be the number of cameras and nE be the number of edges.
%
% INPUT:
% A is the nV x nV (symmetric) adjacency matrix associated to a connected
% graph: A(i,j)=1 if the edge (i,j) appears in the graph, A(i,j) = 0
% otherwise.
%
% [optional] cam is the root of the spanning tree. If not specifid then a
% random  index between 1 and nV is chosen.
%
% OUTPUT:
% C is a nE-nV+1 x nE matrix which contains a cycle basis for the
% epipolar graph viewed as a vector space, i.e. C(i,:) is the signed
% indicator vector associated to the i-th cycle in the basis.

% Algorithm: The spanning tree is computed as a maximum-weight spanning
% tree with a given root by Kruskal's algorithm, so as to allow the
% generation of multiple random spanning trees. The time complexity of
% Kruskal's algorithm is O(nE+x*log(nV)) where x is the number of edges no
% longer than the longest edge in the spanning tree.
%
% Authors: Federica Arrigoni, Beatrice Rossi, Andrea Fusiello, 2015


global debug;

method = 'MST';

if nargin < 2;
    bc = betweenness_centrality(W);
    [~,root]=max(bc);
end

W(logical(eye(size(W)))) = 0; % azzera diagonale di W
W(W>0) = 1./(W(W>0)+eps);
nE=nnz(W); % number of edges

if debug
    % check if A is connected
    [~,comp]=components(W);
    assert(length(comp)==1)
end

G = graph(W);

clearvars W; %--- W non serve pi

if strcmp(method,'MST') % minimum spanning tree
    [T, pred]  = minspantree(G,'Root', root);
elseif strcmp(method,'SPT') % shortest path tree;
    % In alternativa si pu usare lo shortest path tree; produce cicli pi
    % piccoli, ma non funziona bene
    pred = shortestpathtree(G,root,'OutputForm','vector');
    T = shortestpathtree(G,root);
else
    error('The cycle basis can be MCB | SPT');
end


% edge rimaste fuori
[I,J]=find(tril(adjacency(G)-adjacency(T),-1));

% Per ogni edge non in T crea un ciclo
row=[]; col=[]; val=[];
for c=1:length(I) % c edge da chiudere
    
    v_cycle=zeros(nE,1);
    
    % Find the shortest path in T between I(c) and J(c)
    [k,path] = lca(I(c),J(c),pred);
    v_index = findedge(G,path,circshift(path,[0,-1]));
    
    % Compute the cycle generated by T and the edge I(c),J(c)
    for k=2:length(path)
        u=path(k-1); v= path(k);
        v_cycle(v_index(k-1)) = 1.*(u>v)-1.*(u<v);
    end
    % chiusura path
    u=path(k); v=path(1);
    v_cycle(v_index(end)) = 1.*(u>v)-1.*(u<v);
    
    if debug
        assert(length(path) > 2);
    end
    
    foo = find(v_cycle);
    row  = [row; foo];
    col  = [col; c*ones(size(foo)) ] ;
    val =  [val; v_cycle(foo)];
end

C = sparse(col,row,val);

%max(sum(abs(C),2))

end


function [k,path] = lca(i,j,pred)
% find  least common ancestor of i and j in the tree encoded  by pred
% and the path from i to j passing through k


% percorso da i alla radice
pi=[];
while i~=0,
    pi(end+1)=i;
    i=pred(i);
end;

% percorso da j alla radice
pj=[];
while j~=0,
    pj(end+1)=j;
    j=pred(j);
end;

% trova least common ancestor
[~,LOCB] = ismember(pi,pj);
LOCB(LOCB==0) = [];
k = pj(min(LOCB));
path = [ pi(1:find(pi==k)-1), k,  fliplr(pj(1:find(pj==k)-1))];

end


